from rest_framework.renderers import JSONRenderer

from .utils import get_param


class DatatablesRenderer(JSONRenderer):
    media_type = 'application/json'
    format = 'datatables'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into JSON, returning a bytestring.
        """
        if data is None:
            return bytes()

        request = renderer_context['request']
        new_data = {}

        view = renderer_context.get('view')

        if 'recordsTotal' not in data:
            # pagination was not used, let's fix the data dict
            if 'results' in data:
                results = data['results']
                count = data['count'] if 'count' in data else len(results)
            else:
                results = data
                count = len(results)
            new_data['data'] = results
            if view and hasattr(view, '_datatables_filtered_count'):
                count = view._datatables_filtered_count
            if view and hasattr(view, '_datatables_total_count'):
                total_count = view._datatables_total_count
            else:
                total_count = count
            new_data['recordsFiltered'] = count
            new_data['recordsTotal'] = total_count
        else:
            new_data = data
        # add datatables "draw" parameter
        new_data['draw'] = int(get_param(request, 'draw', '1'))

        serializer_class = None
        if hasattr(view, 'get_serializer_class'):
            serializer_class = view.get_serializer_class()
        elif hasattr(view, 'serializer_class'):
            serializer_class = view.serializer_class

        if serializer_class is not None and hasattr(serializer_class, 'Meta'):
            force_serialize = getattr(
                serializer_class.Meta, 'datatables_always_serialize', ()
            )
        else:
            force_serialize = ()

        self._filter_unused_fields(request, new_data, force_serialize)

        if hasattr(view.__class__, 'Meta'):
            extra_json_funcs = getattr(
                view.__class__.Meta, 'datatables_extra_json', ()
            )
        else:
            extra_json_funcs = ()

        self._filter_extra_json(view, new_data, extra_json_funcs)

        return super(DatatablesRenderer, self).render(
            new_data, accepted_media_type, renderer_context
        )

    def _filter_unused_fields(self, request, result, force_serialize):
        # list of params to keep, triggered by ?keep= and can be comma
        # separated.
        keep = request.query_params.get('keep', [])
        cols = []
        i = 0
        while True:
            col = get_param(request, 'columns[%d][data]' % i)
            if col is None:
                break
            cols.append(col.split('.').pop(0))
            i += 1
        if len(cols):
            data = result['data']
            for i, item in enumerate(data):
                try:
                    keys = set(item.keys())
                except AttributeError:
                    continue
                for k in keys:
                    if (
                            k not in cols
                            and not k.startswith('DT_Row')
                            and k not in force_serialize
                            and k not in keep
                    ):
                        result['data'][i].pop(k)

    def _filter_extra_json(self, view, result, extra_json_funcs):
        read_only_keys = result.keys()  # don't alter anything
        for func in extra_json_funcs:
            if not hasattr(view, func):
                raise TypeError(
                    "extra_json_funcs: {0} not a view method.".format(func)
                )
            method = getattr(view, func)
            if not callable(method):
                raise TypeError(
                    "extra_json_funcs: {0} not callable.".format(func)
                )
            key, val = method()
            if key in read_only_keys:
                raise ValueError("Duplicate key found: {key}".format(key=key))
            result[key] = val



from django.db.models import Q


class SwitchRegexFilter(object):

    lookup_regex_replacements = {
      'icontains': 'iregex',
      'contains': 'regex',
      'exact': 'regex',
    }

    def __init__(self, field_name=None, lookup_expr='exact', *, label=None,
                 method=None, distinct=False, exclude=False, **kwargs):
        self._original_lookup_expr = lookup_expr
        super().__init__(
            field_name=field_name, lookup_expr=lookup_expr, label=label,
            method=method, distinct=distinct, exclude=exclude, **kwargs)

    @classmethod
    def replace_last_lookup(cls, lookup_expr, replacement=None):
        lookups = lookup_expr.split('__')
        last_lookup = lookups[-1]
        if replacement is None:
            replacement = cls.lookup_regex_replacements.get(
                last_lookup, last_lookup)
        lookups[-1] = replacement
        return '__'.join(lookups)

    def lookup_expr():
        def fget(self):
            if not self.search_regex:
                return self._original_lookup_expr
            else:
                return self.replace_last_lookup(self._original_lookup_expr)

        def fset(self, value):
            self._original_lookup_expr = value
        return locals()

    lookup_expr = property(**lookup_expr())

    @property
    def search_regex(self):
        # datatables_query is only present, if there's a query for
        # this column, so even if we use the correct backend, it might
        # not be.
        return getattr(self, 'datatables_query', {}).get('search_regex')


class GlobalFilter(SwitchRegexFilter):
    """Simple global filter mixin that duplicates the behaviour of the
    global filtering without using django-filter.

    *Must* be used with DjangoFilterBackend which combines the global
    and colum searches in the correct way and with
    DatatablesFilterSet, which ensures that the correct attributes are
    set to make this work.

    For more fine-grained control over global filter behaviour, you
    should implement the global_q method yourself.

    Examples
    --------

    >>> class GlobalCharFilter(GlobalFilter, filters.CharFilter):
    ...     pass
    ...

    >>> class AlbumGlobalFilter(AlbumFilter):
    ...     name = GlobalCharFilter(lookup_expr='icontains')
    ...     genres = GlobalCharFilter(lookup_expr='name__icontains')
    ...     artist = GlobalCharFilter(lookup_expr='name__icontains')
    ...     year = GlobalCharFilter()
    ...     class Meta:
    ...         model = Album
    ...         fields = '__all__'
    ...

    see tests/test_django_filter_backend.py

    """

    def global_q(self):
        """Return a Q-Object for the local search for this column"""
        ret = Q()
        if self.global_search_value:
            ret = Q(**{self.global_lookup: self.global_search_value})
        return ret

    @property
    def global_lookup(self):
        return ('{}__{}'
                .format(self.field_name,
                        self.replace_last_lookup(self.lookup_expr,
                                                 self.global_lookup_expr)))

    @property
    def global_lookup_expr(self):
        if self.global_search_regex:
            return 'iregex'
        return 'icontains'

    @property
    def global_search_regex(self):
        assert hasattr(self, '_global_search_regex'), (
            'Must be used with e.g. DatatablesFilterSet to ensure '
            '_global_search_regex is set')
        return self._global_search_regex

    @property
    def global_search_value(self):
        assert hasattr(self, '_global_search_value'), (
            'Must be used with e.g. DatatablesFilterSet to ensure '
            '_global_search_value is set')
        return self._global_search_value



from django.db.models import Q
from django_filters.rest_framework.backends import DjangoFilterBackend
from django_filters import utils

from rest_framework_datatables import filters

from .filterset import DatatablesFilterSet


class DatatablesFilterBackend(filters.DatatablesBaseFilterBackend,
                              DjangoFilterBackend):

    filterset_base = DatatablesFilterSet

    def filter_queryset(self, request, queryset, view):
        """Filter DataTables queries with a filterset

        This method needs to combine the necessary steps from both its
        superclasses.

        """
        if not self.check_renderer_format(request):
            return queryset

        self.set_count_before(view, view.get_queryset().count())

        # parsed datatables_query will be an attribute of the filterset
        filterset = self.get_filterset(request, queryset, view)
        if filterset is None:
            self.set_count_after(view, queryset.count())
            return queryset

        if not filterset.is_valid() and self.raise_exception:
            raise utils.translate_validation(filterset.errors)
        queryset = filterset.qs
        global_q = self.get_global_q(filterset)
        if global_q:
            queryset = queryset.filter(global_q).distinct()

        self.set_count_after(view, queryset.count())

        # TODO Can we use OrderingFilter, maybe in DatatablesFilterSet, by
        # default? See
        # https://django-filter.readthedocs.io/en/master/ref/filters.html#ordering-filter
        ordering = self.get_ordering(request, view, filterset)
        if ordering:
            queryset = queryset.order_by(*ordering)

        return queryset

    def get_filterset_kwargs(self, request, queryset, view):
        query = self.parse_datatables_query(request, view)
        return {
            'data': query['form_fields'],
            'queryset': queryset,
            'request': request,
            'datatables_query': query
        }

    def parse_datatables_query(self, request, view):
        query = super().parse_datatables_query(request, view)
        form_fields = {}
        field_queries = {}
        for f in query['fields']:
            form_fields[f['data']] = f['search_value']
            field_queries[f['data']] = f
        query['form_fields'] = form_fields
        query['field_queries'] = field_queries
        return query

    def get_global_q(self, filterset):
        global_q = Q()
        for filter_name, f in filterset.filters.items():
            if (
                    filter_name in filterset.datatables_query['field_queries']
                    and hasattr(f, 'global_q')):
                global_q |= f.global_q()
        return global_q

    def get_ordering(self, request, view, filterset):
        ret = []
        for field, dir_ in self.get_ordering_fields(
                request, view, filterset.datatables_query['fields']):
            if field['data'] in filterset.filters:
                filter = filterset.filters[field['data']]
                lookup = '__'.join(
                    f'{filter.field_name}__{filter.lookup_expr}'
                    .split('__')
                    [:-1])
                ret.append(('-' if dir_ == 'desc' else '')
                           + lookup)
        self.append_additional_ordering(ret, view)
        return ret
